# ZShellScripts - (c) 2002 Jerome Alet
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# author: Jerome Alet - <alet@librelogiciel.com>
#
# $Id: zshell.py,v 1.2 2002/11/20 12:06:10 jalet Exp $
#
# $Log: zshell.py,v $
# Revision 1.2  2002/11/20 12:06:10  jalet
# Bug with Zope 2.6 and Python 2.2.2 wrt sys.hexversion, solved with
# the help of Guido van Rossum.
#
# Revision 1.1  2002/11/05 10:24:13  jalet
# Preliminary version of stripped down ZShell added.
#
#
#

# standard modules
import sys
import os
import copy
import string
import re
import urllib
import cStringIO
import posixpath
import shlex
import rexec
import getopt
import fnmatch
import time
import popen2
import glob

# Zope modules
import Globals
import AccessControl, AccessControl.SecurityManagement
from OFS.DTMLDocument import DTMLDocument
from OFS.DTMLMethod import DTMLMethod
from OFS.CopySupport import CopyError
from Products.PythonScripts.PythonScript import PythonScript
from DateTime import DateTime
from ZODB import FileStorage

class ZopeShell :
        # Just to be sure
        __context = None
        __version = None
        __status = None
        __stdin = None
        __stdout = None
        __stderr = None
        __htmlfinal = None
        __temphtml = None
        __olderror = None

        def __init__(self, context, zshellscript = None, env = os.environ) :
                # the current working Folder may be stored in
                # an hidden field in the form: check if it is or not
                # and set private context accordingly
                if hasattr(context, "REQUEST") and context.REQUEST.has_key("zshellpwd") :
                        pwdobj = self.toObject(context, context.REQUEST["zshellpwd"])
                        if pwdobj is not None :
                                self.__context = pwdobj
                        else :
                                self.__context = context
                else :
                        self.__context = context

                # get active version, if any
                if self.__context.REQUEST.cookies.has_key(Globals.VersionNameName) :
                        self.__version = self.__context.REQUEST.cookies[Globals.VersionNameName]
                else :
                        self.__version = None

                # set initial status to None, meaning no running command
                self.__status = None

                # now we open the standard IO channels
                self.__stdin = cStringIO.StringIO()
                self.__stdout = cStringIO.StringIO()
                self.__stderr = cStringIO.StringIO()

                # create an empty HTML document for final result
                self.__htmlfinal = jaxml.HTML_document()
                self.__htmlfinal._indentstring(' ')

                # create an empty HTML document for commands results
                self.__temphtml = jaxml.HTML_document()
                self.__temphtml._indentstring(' ')

                # put intial content in the html final output
                self.initHTML()

                # get initial environment variables
                self.__env = copy.deepcopy(env)

                # then add the REQUEST object's content
                for k in self.__context.REQUEST.keys() :
                        self.__env["%s" % k] = str(self.__context.REQUEST[k])

                # get commands either from parameters or from REQUEST
                commands = self.getCommands(zshellscript)

                # we execute commands before displaying the prompt
                # to be able to take care of a change of the current working directory.
                # but we only execute them if zshelldontrun is not in the REQUEST, because
                # this would mean that probably the user just wants the help
                # and the insertion of an empty command in the textarea
                if not self.__context.REQUEST.has_key("zshelldontrun") :
                        status = self.execCommands(commands)
                else :
                        # only executes the man commands, ignore the rest
                        status = self.execCommands(filter(lambda f: f[:4] == "man ", commands))

                # displays the prompt
                self.showPrompt(commands)

        def getContext(self) :
                return self.__context

        def getStatus(self) :
                return self.__status

        def errormessage(self, msg) :
                """Outputs an error message to the pseudo stderr stream.
                
                   don't output anything if the previous message was the same.
                """
                if msg != self.__olderror :
                        self.__stderr.write("%s\n" % msg)
                self.__olderror = msg[:]
                return -1

        def permissionProblem(self, object, perm) :
                msg = "You don't have permission '%s' on"
                path = self.ObjectPath(object)
                if path is not None :
                        return self.errormessage((msg + " %s") % (perm, path))
                else :
                        return self.errormessage((msg + " object %s") % (perm, repr(object)))

        def roleProblem(self, object, role) :
                msg = "You don't have role '%s' on"
                path = self.ObjectPath(object)
                if path is not None :
                        return self.errormessage((msg + " %s") % (role, path))
                else :
                        return self.errormessage((msg + " object %s") % (role, repr(object)))

        def htmlstr(self, unsafe) :
                """Formats a string to be shown safely in HTML"""
                unsafe = string.replace(unsafe, '&', '&amp;')
                unsafe = string.replace(unsafe, '<', '&lt;')
                return string.replace(unsafe, '>', '&gt;')

        def printf(self, msg) :
                """outputs a message only to stdout"""
                self.__stdout.write("%s" % msg)

        def ShellExpand(self, path, zodb=1) :
                """Simulates a shell expansion using glob.glob and redirecting some of the os module functions to our own"""
                def mylistdir(path, zopeshell = self) :
                        # the original os.listdir accepts only one argument, but we need
                        # both the path to scan, and a reference to the current object,
                        # so we "cheat" using a default argument which is never used
                        # by os.listdir.
                        obj = zopeshell.toObject(zopeshell.getContext(), path)
                        if (obj is not None) and hasattr(obj, "isPrincipiaFolderish") \
                            and obj.isPrincipiaFolderish and hasattr(obj, "objectIds") :
                                if zopeshell.HasPerms(obj, 'Access contents information', verbose = 0) :
                                        return obj.objectIds()
                        return []

                def myexists(path, zopeshell = self) :
                        # see above
                        return zopeshell.toObject(zopeshell.getContext(), path) is not None

                def myisdir(path, zopeshell = self) :
                        # see above
                        while path and (path[-1] == '/') :
                                path = path[:-1]
                        obj = zopeshell.toObject(zopeshell.getContext(), path or '/')
                        if (obj is not None) and hasattr(obj, "isPrincipiaFolderish") :
                                return obj.isPrincipiaFolderish
                        else :
                                return 0

                # builds a restricted execution environment.
                # this way we are thread safe. Thanks to Toby Dickenson.
                r_env = rexec.RExec()
                osmodule = r_env.add_module('os')
                for name,value in os.__dict__.items():
                    setattr(osmodule,name,value)
                ospathmodule = r_env.add_module('os.path')
                for name,value in os.path.__dict__.items():
                    setattr(ospathmodule,name,value)

                if zodb :
                        # we want to expand wildcards from the ZODB
                        # so we plug in our own methods
                        osmodule.listdir = mylistdir
                        ospathmodule.join = posixpath.join
                        ospathmodule.split = posixpath.split
                        ospathmodule.exists = myexists
                        ospathmodule.isdir = myisdir

                # do globbing using our own methods
                # or the normal ones if zodb==0
                # python 2.2 bug in sre wrt sys.hexversion
                # solved with the help of Guido van Rossum !
                r_env.s_exec('import sys\nsys.hexversion = 0\nimport glob\n')
                result = r_env.r_eval('glob.glob')(path)
                if not result :
                        result = [path] # returns the argument unmodified
                return result

        def ObjectPath(self, object) :
                if hasattr(object, "absolute_url") :
                        url = object.absolute_url()
                        return '/' + string.join(string.split(url, '/')[3:], '/')

        def getcwd(self):
                return self.ObjectPath(self.__context)

        def toObject(self, curdir, path) :
                # Michel@DC: ouch, it looks like you go about doing your own
                # traversal over Zope objects here, this is bad.
                # because of many layers of history, traversal is a
                # delicate process that you will probably get wrong if
                # you try to do it yourself.  Instead of doing your
                # own traversal, use
                # self.__context.restrictedTraverse() to do the
                # traversal for you (assuming __context is a Zope
                # container) as a bonus, restrictedTraverse will
                # enforce security for you.

                # Jerome: Here's my try, but with unrestrictedTraverse.
                # it seems that restrictedTraverse is a bit, hmmm..., too restricted !
                path = posixpath.normpath(path)
                try :
                        if path == '.' :
                                return self.__context
                        elif path == '..' and hasattr(self.__context,'aq_parent') :
                                if hasattr(self.__context, "isTopLevelPrincipiaApplicationObject") and self.__context.isTopLevelPrincipiaApplicationObject :
                                        return self.__context
                                else :
                                        return self.__context.aq_parent
                        else :
                                return curdir.unrestrictedTraverse(path)
                except IndexError:
                        # BUG in Zope 2.3.2 ! => doesn't take care correctly of the / folder
                        # when the path argument contains only '/'
                        return curdir.getPhysicalRoot()
                except AttributeError:
                        return None     # nothing, e.g. acl_users/*
                except KeyError:
                        return None     # empty folder it seems
                except TypeError:
                        return None     # What the hell is this ?

        def getCommands(self, script = None) :
                if script is None :
                        # no argument, get them from REQUEST
                        if self.__context.REQUEST.has_key("zshellscript") :
                                script = self.__context.REQUEST["zshellscript"]
                        else :
                                script = []
                if type(script) != type([]) :
                        script = [script]
                return map(string.strip, filter(None, script))

        def execCommands(self, cmds) :
                self.__status = 0
                if cmds :
                        for command in cmds :
                                self.__status += self.execCommand(command) or 0
                        if self.__status :
                                self.__stderr.write("WARNING: %i errors were encountered\n" % abs(self.__status))
                return self.__status

        def execCommand(self, cmdline) :
                # Unfortunately shlex.shlex needs a file object, not a buffer.
                # tokenize the command line
                fcmdline = cStringIO.StringIO("%s" % cmdline)
                tokenizer = shlex.shlex(fcmdline)
                tokenizer.wordchars = tokenizer.wordchars + r".:,?!~/\_$*-+={}[]()"
                tokenizer.quotes = tokenizer.quotes + "`"
                unexpanded = []
                while 1 :
                        token = tokenizer.get_token()
                        if token :
                                unexpanded.append(token)
                        else :
                                break
                fcmdline.close()
                del tokenizer
                del fcmdline

                # split the arguments list on ';' chars
                # to allow multiple commands separated by ';'
                splitcommands = []
                try :
                        while unexpanded :
                                i = unexpanded.index(';')
                                current, unexpanded = unexpanded[:i], unexpanded[i+1:]
                                splitcommands.append(current)
                except ValueError :
                        splitcommands.append(unexpanded)

                status = 0
                for unexpanded in splitcommands :
                        if not unexpanded :
                                continue
                        # the command is the first element
                        cmd = unexpanded[0]
                        cmdname = "run_" + cmd
                        if not hasattr(self, cmdname) :
                                # TODO : horrible hack here
                                # unknown command, try a VAR=VALUE maybe...
                                return self.run_setenv([cmd], [cmd])
                        else :
                                # skip the command
                                unexpanded = unexpanded[1:]
                                # then build a wildcards expanded list of arguments
                                expanded = []
                                for i in range(len(unexpanded)) :
                                        # if quoted, then add it without the quotes
                                        # and modify the unexpanded arg accordingly
                                        uarg = unexpanded[i]
                                        if (uarg[0] == uarg[-1] == "'") \
                                           or (uarg[0] == uarg[-1] == '"') :
                                                arg = self.expandvar(uarg[1:-1])
                                                expanded.append(arg)
                                                unexpanded[i] = arg
                                        else :
                                                # if not quoted then try to expand wildcards
                                                # and add the result
                                                arg = self.expandvar(uarg)
                                                expanded.extend(self.ShellExpand(arg) or [arg])

                                # sets some additionnal environment variables
                                self.setadditionalvars([ cmdname[4:] ] + expanded)

                                # then launch the command: we don't do it before updating
                                # the history because we would miss shutdowns and restarts
                                status += getattr(self, cmdname)(expanded, unexpanded) or 0

                                # unsets the additional environment variables
                                self.unsetadditionalvars([ cmdname[4:] ] + expanded)

                return status

        def unsetadditionalvars(self, values) :
                """Unsets additional environment variables, ie $0 to $n ..."""
                for i in range(len(values)) :
                        key = "%i" % i
                        if self.__env.has_key(key) :
                                del self.__env[key]
                # horrible hack :
                if self.__env.has_key("*") :
                        del self.__env["*"]

        def setadditionalvars(self, values) :
                """Sets additional environment variables, ie $0 to $n ..."""
                for i in range(len(values)) :
                        self.__env["%i" % i] = values[i]
                # horrible hack :
                if values :
                        self.__env["*"] = string.join(values[1:], " ")

        def expandvar(self, var) :
                """Expands a variable (which name begins with a dollar sign) to its value."""
                if (len(var) > 1) and (var[0] == '$') :
                        # TODO : make it smarter wrt computations like $[$VARNAME] and so on...
                        return self.__env.get(var[1:], "")
                else :
                        return var

        def HasPerms(self, object, perms, verbose = 1) :
                """Checks if the user has all the perms permissions on an object"""
                if type(perms) != type([]) :
                        perms = [perms]
                SecurityManager = AccessControl.getSecurityManager()
                for perm in perms :
                        if not SecurityManager.checkPermission(perm, object) :
                                if verbose :
                                        self.permissionProblem(object, perm)
                                return 0
                return 1

        def HasRoles(self, object, roles, verbose = 1) :
                """Checks if the user has at least one role in roles in the context of an object"""
                if type(roles) != type([]) :
                        roles = [roles]
                ctxtroles = AccessControl.getSecurityManager().getUser().getRolesInContext(object)
                for role in roles :
                        if role in ctxtroles :
                                return 1
                if verbose :
                        self.roleProblem(object, role)
                return 0

        def WhoAmI(self) :
                """Returns the current user's name"""
                user = AccessControl.getSecurityManager().getUser()
                return (user.getUserName(), user.getRoles())

        def getMetaType(self, object) :
                """Returns an object's meta type"""
                if callable(object.meta_type) :
                        # at least ZClasses
                        return object.meta_type()
                else :
                        # the rest
                        return object.meta_type

        def getArgsAndNot(self, allargs) :
                """Splits a list of arguments into two sublists:
                        - a list of arguments
                        - a list of NOTarguments
                   a NOTargument is an argument which begins with '!'
                """
                args = []
                notargs = []
                for arg in allargs :
                        if arg[0] == '!' :
                                # add it to the notargs list but we
                                # don't want the '!' character
                                notargs.append(arg[1:])
                        else :
                                args.append(arg)
                if notargs and not args :
                        # args still empty, we want all of them
                        # before matching negatively
                        args = ["*"]
                return (args, notargs)

        def descend(self, root, func, maxdepth = 0, curdepth = 0) :
                status = 0
                if root is not None :
                        if not self.HasPerms(root, 'View') :
                                status = -1
                        else :
                                status = status + func(root)
                                if (not (maxdepth and (curdepth >= maxdepth))) and hasattr(root, "isPrincipiaFolderish") and root.isPrincipiaFolderish :
                                        curdepth = curdepth + 1
                                        if not self.HasPerms(root, 'Access contents information') :
                                                status = status - 1
                                        for object in root.objectValues() :
                                                status = status + self.descend(object, func, maxdepth, curdepth)
                return status

        def getopt(self, longs, argv) :
                """Analyses command line options, only long options are recognized"""
                # analyse the arguments to extract those which may be lists
                # from those which must be single values
                # we put them into lists, since not so many options
                # will generally be used, this should be at least as fast
                # as putting them in mappings
                single = []
                multiple = []
                for l in range(len(longs)) :
                        long = longs[l]
                        if long[-1] == '+' :
                                # this option may be specified multiple times
                                # but we must modify it for getopt.getopt to
                                # work correctly
                                longs[l] = long[:-1] + '='
                                multiple.append(long[:-1])
                        elif long[-1] == '=' :
                                single.append(long[:-1])
                        else :
                                single.append(long)
                try :
                        result = {}
                        options,args = getopt.getopt(argv, '', longs)
                        if options :
                                for (o, v) in options :
                                        o = o[2:]
                                        if o in single :
                                                if not result.has_key(o) :
                                                        result[o] = v
                                                else :
                                                        raise getopt.error, "Option --%s can't be specified more than one time on this command line" % o
                                        elif o in multiple :
                                                if not result.has_key(o) :
                                                        result[o] = []
                                                result[o].append(v)
                                        else :
                                                # there's a very big problem !
                                                raise getopt.error, "ZShell internal error while parsing command line arguments"
                        elif not args :
                                args = argv     # no option and no argument, return argv inchanged
                        return (result, args)
                except getopt.error, msg :
                        self.errormessage("%s" % msg)
                        return (None, None)

        def match_anystring(self, option, value, options) :
                if options.has_key(option) :
                        if value :
                                (vals, notvals) = self.getArgsAndNot(options[option])
                                ok = 0
                                for val in vals :
                                        if fnmatch.fnmatchcase(value, val) :
                                                ok = 1
                                                break
                                oknot = 0
                                for notval in notvals :
                                        if fnmatch.fnmatchcase(value, notval) :
                                                oknot = 1
                                                break
                                return ok and (not oknot)
                        return 0
                return 1

        def match_MetaType(self, object, options) :
                """Returns 1 if an object meta type matches optional --type options.
                   If no --type option is given, then the object always matches.
                """
                return self.match_anystring("type", self.getMetaType(object), options)

        def match_Id(self, object, options) :
                """Returns 1 if an object's id matches optional --id options.
                   If no --id option is given, then the object always matches.
                """
                return self.match_anystring("id", object.getId(), options)

        def match_Owner(self, object, options) :
                """Returns 1 if an object's owner matches optional --owner options.
                   If no --owner option is given, then the object always matches.
                """
                if options.has_key("owner") :
                        ownerinfo = object.owner_info()
                        if ownerinfo is not None :
                                # at least /Control_Panel/Products[/*] doesn't satisfy the following test
                                if hasattr(ownerinfo, "has_key") and ownerinfo.has_key('id') :
                                        return self.match_anystring("owner", ownerinfo['id'], options)
                        return 0 # Not owned or no owner: never match
                return 1         # No owner option: always match

        def match_Newer(self, object, options) :
                """Returns 1 if an object is newer than an optional --newer option.
                   Returns -1 if the --newer option argument doesn't exist
                   Returns 0 if the object is not newer
                   if modification times are equal then the object matches.
                   If no --newer option is given, then the object always matches.
                """
                if options.has_key("newer") :
                        objnewer = self.toObject(self.__context, options["newer"])
                        if objnewer is not None :
                                if object.bobobase_modification_time() <= objnewer.bobobase_modification_time() :
                                        return 0
                        else :
                                return self.errormessage("Object %s doesn't exist" % options["newer"])
                return 1

        def match_Older(self, object, options) :
                """Returns 1 if an object is older than an optional --older option.
                   Returns -1 if the --older option argument doesn't exist
                   Returns 0 if the object is not older
                   if modification times are equal then the object matches.
                   If no --older option is given, then the object always matches.
                """
                if options.has_key("older") :
                        objolder = self.toObject(self.__context, options["older"])
                        if objolder is not None :
                                if object.bobobase_modification_time() >= objolder.bobobase_modification_time() :
                                        return 0
                        else :
                                return self.errormessage("Object %s doesn't exist" % options["older"])
                return 1

        def match_Time(self, object, options) :
                """Returns 1 if an object is newer than an optional --mmin or --mtime option.
                   Returns -1 if the --mmin or --mtime option argument is an invalid value.
                   --mmin expects a duration in minutes.
                   --mtime expects a duration in days.
                   Returns 0 if the object is not newer.
                   If none of --mmin or --mtime option is given, then the object always matches.
                """
                if options.has_key("mmin") or options.has_key("mtime") :
                        if options.has_key("mmin") and options.has_key("mtime") :
                                return self.errormessage("Options --mmin and --mtime are incompatible")
                        if options.has_key("mmin") :
                                timestr = options["mmin"]
                        else :
                                timestr = options["mtime"]
                        try :
                                if options.has_key("mmin") :
                                        modtime = int(timestr) * 60
                                else :
                                        modtime = int(timestr) * 60 * 60 * 24
                        except ValueError :
                                return self.errormessage("Invalid time %s" % timestr)
                        testtime = DateTime(int(time.time()) - modtime)
                        if object.bobobase_modification_time() <= testtime :
                                return 0
                return 1

        def match_Many(self, object, options) :
                """Returns 1 if an object matches all criterias.
                   Returns 0 if it doesn't match some of them.
                   Returns -1 if there's an error somewhere.
                """
                # if id doesn't match then skip
                if not self.match_Id(object, options) :
                        return 0

                # if meta type doesn't match then skip
                if not self.match_MetaType(object, options) :
                        return 0

                # if owner doesn't match then skip
                if not self.match_Owner(object, options) :
                        return 0

                # if not newer than another object then skip
                newer = self.match_Newer(object, options)
                if newer <= 0 :
                        return newer

                # if not older than another object then skip
                older = self.match_Older(object, options)
                if older <= 0 :
                        return older

                # if modification time doesn't match options
                # --mmin or --mtime then skip
                matchtime = self.match_Time(object, options)
                if matchtime <= 0 :
                        return matchtime
                return 1

        def getMaxDepth(self, options, default=0) :
                mx = default
                if options.has_key("maxdepth") :
                        try :
                                mx = int(options["maxdepth"])
                        except ValueError :
                                return self.errormessage("the --maxdepth option's argument must be a numeric")
                return mx

        def mv_or_cp(self, cmd, expanded) :
                if len(expanded) < 2 :
                        return self.errormessage("Incorrect number of arguments")
                else :
                        status = 0
                        dst = expanded[-1]
                        srcs = expanded[:-1]
                        objids = []
                        for src in srcs :
                                if '/' in src :
                                        status = status + self.errormessage('Paths in source objects are not allowed at this time: %s' % src)
                                else :
                                        objids.append(src)
                        dsto = self.toObject(self.__context, dst)
                        if dsto is None :
                                return status + self.errormessage("Incorrect destination argument: %s" % dst)
                        if not dsto.isPrincipiaFolderish :
                                return status + self.errormessage("Not a folderish object: %s" % dst)

                        # Michel@DC: here you should do a
                        # SecurityManager.checkPermission('View
                        # management screens', self.__context) to make
                        # sure the user has permission to copy or
                        # paste.

                        # Jerome: In fact we must test this perm on both
                        # the source and the destination.
                        if not self.HasPerms(self.__context, 'View management screens') :
                                return status - 1
                        if not self.HasPerms(dsto, 'View management screens') :
                                return status - 1

                        # All is fine, do it now.
                        try :
                                if cmd == 'cp' :
                                        self._clipboard = self.__context.manage_copyObjects(ids = objids)
                                        action = 'copied'
                                else :
                                        self._clipboard = self.__context.manage_cutObjects(ids = objids)
                                        action = 'moved'
                                dsto.manage_pasteObjects(cb_copy_data = self._clipboard)
                                for oid in objids :
                                        self.htmlmessage('%s %s to %s' % (oid, action, self.ObjectPath(dsto)))
                        except AttributeError, msg:
                                status = status + self.errormessage("Object %s doesn't exist" % msg)
                        except CopyError :
                                status = status + self.errormessage("Objects can't be %s to %s" % (action, self.ObjectPath(dsto)))
                        return status

        def run_exec(self, expanded, unexpanded) :
                """Executes an external command

                   exec ls -la /

                   Hint: the external command is launched as the user
                         who launched Zope (usually user nobody)

                   Caveats: things like 'telnet' need more work ;-)

                   WARNING: like with the find command, you will become
                            addicted quick.
                """
                process = popen2.Popen3(string.join(unexpanded, " "), capturestderr=1)
                retcode = process.wait()
                output = process.fromchild.read()
                errors = process.childerr.read()
                if errors :
                        # self.errormessage already outputs a final \n
                        if errors[-1] == '\n' :
                                errors = errors[:-1]
                        self.errormessage(errors)
                self.htmlmessage(string.replace(output, '\n', '<BR>'), safe=1, printable=1)
                del process
                if errors :
                        return -1

        # Michel@DC: all methods below here need to validate their operations
        # with SecurityManager.checkPermission

        def run_su(self, expanded, unexpanded) :
                """Run a command as another user

                   Accepts a --user xxxx option
                   and a --password yyyy option.

                   If the current user has a Manager role in the current context,
                   then no password is required, else the correct password for
                   user xxxx must be entered.
                   If there's no --user option, then an su to user 'admin' is tried:

                   su --user jerome --password x./32 "rm /QuickStart"

                   Nota Bene: Both the password and domains must validate for
                              the new user.

                   Caveats: su state is not preserved, it's done voluntarily
                            this way, but just tell me if you prefer another behavior.
                """
                options, args = self.getopt(["user=", "password="], unexpanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                if not args :
                        return self.errormessage("Needs a command to run as another user")
                # we display the same message in all cases to prevent
                # a brute force attack to learn existing usernames
                incorrect_user_or_password = "You must supply a correct username and password"
                if not self.HasRoles(self.__context, 'Manager', verbose=0) :
                        # not an admin, password is required
                        if not options.has_key("password") :
                                return self.errormessage("%s" % incorrect_user_or_password)
                        else :
                                password = options["password"]
                else :
                        password = None
                if not options.has_key("user") :
                        newusername = "admin"
                else :
                        newusername = options["user"]

                newuser = self.__context.acl_users.getUser(newusername)
                if (newuser is None) or ((password is not None) and not (newuser.authenticate(password, self.__context.REQUEST))) :
                        return self.errormessage("%s" % incorrect_user_or_password)
                olduser = AccessControl.getSecurityManager().getUser()
                oldusername = olduser.getUserName()
                AccessControl.SecurityManagement.newSecurityManager(None, newuser)
                self.htmlmessage("User '%s' does a su to '%s'" % (oldusername, newusername))
                status = self.execCommand(string.join(args, ' '))
                if oldusername == 'Anonymous User' :
                        AccessControl.SecurityManagement.noSecurityManager()
                else :
                        AccessControl.SecurityManagement.newSecurityManager(None, olduser)
                self.htmlmessage("Current user '%s' was reset to '%s'" % (newusername, oldusername))
                return status

        def run_grep(self, expanded, unexpanded) :
                """Search and optionally replace regexps in objects contents

                   Accepts many options:

                        --recurse       recursive search
                        --maxdepth n    descend at most n levels
                        --type metatype search only objects of meta type xxx
                        --owner xxxx    objects owned by user xxxx
                        --newer /other/object   objects modified more recently than other
                        --older /other/object   objects modified less recently than other
                        --mmin n        objects modified less than n minutes ago
                        --mtime n       objects modified less than n days ago
                        --properties    search in properties too
                        --replace xxxx  replaces each occurence of the pattern
                                        with the string xxx.
                        --invert        invert the search: objects that don't match
                                                           the regexp are selected
                        --ignorecase    ignore case: this is just for convenience since
                                        it's always possible to use a pattern which
                                        explicitly asks for ignoring case.

                   e.g.:

                     grep --recurse --ignorecase ^word" *_html

                     this one will search recursively for each document which
                     id ends in "_html" and which has "word" at the beginning of
                     a line, ignoring differences between UPPER and lower case.

                     grep --properties --type "DTML*" --type Folder somestring *

                     This one will search for "somestring" in all objects which
                     meta type is Folder or begins with "DTML" and their
                     properties.

                     The string to search for and the optionally replacing
                     string could be any regular expression defined by the
                     standard re python module.

                   Hint: You may use multiple --type or --owner arguments
                         and each can contain wildcards.

                   WARNING: don't forget to enclose your regexps between
                            single or double quotes or else some regexps
                            won't work as expected.

                   Caveats: If not given the --properties option, only
                            searches in the document_src() and title
                            attributes of objects that have a document_src
                            attribute (if not clear email me).
                """   
                options, args = self.getopt(["recurse", "maxdepth=", "owner+", "mmin=", "mtime=", "newer=", "older=", "type+", "replace=", "properties", "invert", "ignorecase"], expanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                if len(args) < 2 :
                        return self.errormessage("Needs one pattern and one object id to grep")
                if options.has_key("invert") and options.has_key("replace") :
                        return self.errormessage("Options replace and invert are incompatible")
                try :
                        if options.has_key("ignorecase") :
                                pattern = re.compile("%s" % args[0], re.I)
                        else :
                                pattern = re.compile("%s" % args[0])
                except re.error :
                        return self.errormessage("You probably entered an invalid regular expression: %s" % args[0])
                def do_grep(obj, zopeshell = self, options=options, pattern=pattern) :
                        if obj is not None:
                                # tests if the object matches the present options
                                if not zopeshell.match_Many(obj, options) :
                                        return 0

                                url = zopeshell.ObjectPath(obj)
                                # in the following lines the absence of an _updateProperty
                                # attribute indicates an object without properties (e.g. a method)
                                # which indicates an object in which we can't search or replace
                                # in properties.
                                ok = 0
                                if options.has_key("replace") :
                                        if options.has_key("properties") and hasattr(obj, "propertyMap") \
                                           and hasattr(obj, "_updateProperty") :
                                                if not zopeshell.HasPerms(obj, "Access contents information") :
                                                        return 0
                                                for prop in obj.propertyMap() :
                                                        value = str(obj.getProperty(prop["id"]))
                                                        (newstring, number) = pattern.subn(options["replace"], value)
                                                        # we test permissions only when needed:
                                                        # it is not quick, but prevent dummy error messages
                                                        if number and zopeshell.HasPerms(obj, "Manage properties") :
                                                                obj._updateProperty(prop["id"], newstring)
                                                                ok = ok + number

                                        if hasattr(obj, "document_src") :
                                                if zopeshell.HasPerms(obj, "View management screens") :
                                                        src = obj.document_src()
                                                        title = obj.title
                                                        (newsrc, numbersrc) = pattern.subn(options["replace"], src)
                                                        (newtitle, numbertitle) = pattern.subn(options["replace"], title)
                                                        if numbersrc :
                                                                src = newsrc
                                                        if numbertitle :
                                                                title = newtitle
                                                        # we test permissions only when needed:
                                                        # it is not quick, but prevent dummy error messages
                                                        if numbersrc or numbertitle :
                                                                ok = ok + numbersrc + numbertitle
                                                                if (((obj.__class__ == DTMLDocument) or (DTMLDocument in obj.__class__.__bases__)) \
                                                                      and zopeshell.HasPerms(obj, "Change DTML Documents")) \
                                                                   or \
                                                                   (((obj.__class__ == DTMLMethod) or (DTMLMethod in obj.__class__.__bases__)) \
                                                                      and zopeshell.HasPerms(obj, "Change DTML Methods")) :
                                                                        obj.manage_edit(src, title)
                                                                elif (((obj.__class__ == PythonScript) or (PythonScript in obj.__class__.__bases__)) \
                                                                      and zopeshell.HasPerms(obj, "Change Python Scripts")) :
                                                                        obj.write(src)
                                                                else :
                                                                        zopeshell.errormessage("Don't know how to modify object %s" % (url or repr(obj)))
                                else :
                                        if options.has_key("properties") and hasattr(obj, "propertyMap") \
                                           and hasattr(obj, "_updateProperty") :
                                                if not zopeshell.HasPerms(obj, "Access contents information") :
                                                        return 0
                                                for prop in obj.propertyMap() :
                                                        value = str(obj.getProperty(prop["id"]))
                                                        result = pattern.findall(value)
                                                        if result is not None :
                                                                ok = ok + len(result)

                                        if hasattr(obj, "document_src") :
                                                if zopeshell.HasPerms(obj, "View management screens") :
                                                        result = pattern.findall(obj.document_src())
                                                        if result is not None :
                                                                ok = ok + len(result)
                                                        result = pattern.findall(obj.title)
                                                        if result is not None :
                                                                ok = ok + len(result)
                                if url is not None :
                                        if (ok and not options.has_key("invert")) or \
                                           ((not ok) and options.has_key("invert")) :
                                                zopeshell.htmlmessage("%s" % url)
                                                zopeshell.printf("%s\n" % url)
                                else :
                                        return zopeshell.errormessage("Error while accessing to object %s" % repr(obj))
                        return 0
                mx = 1  # default = no recursivity
                if options.has_key("recurse") :
                        mx = self.getMaxDepth(options, default=0)       # default full recursivity
                        if mx < 0 :
                                # an error occured, message already displayed
                                return -1
                elif options.has_key("maxdepth") :
                        return self.errormessage("Option --maxdepth was given, but there's no --recurse option")
                status = 0
                for arg in args[1:] :
                        # current depth is already 1.
                        status = status + self.descend(self.toObject(self.__context, arg), do_grep, maxdepth=mx, curdepth=1)
                return status

        def run_find(self, expanded, unexpanded) :
                """Search objects in the Folder hierarchy

                   Accepts a top Folder from which doing the search as
                   its first argument, and options as its other arguments:

                        --maxdepth n    descend at most n levels
                        --id xxxx       objects which id matchs xxxx
                        --type metatype objects which meta_type is metatype
                        --owner xxxx    objects owned by user xxxx
                        --newer /other/object   objects modified more recently than other
                        --older /other/object   objects modified less recently than other
                        --mmin n        objects modified less than n minutes ago
                        --mtime n       objects modified less than n days ago
                        --exec cmd      exec command for each object found
                                        you can use {} to match the current full object id
                                        and single or double quotes around the command itself.

                   e.g.:

                     find / --owner jerome --id "*_html" --maxdepth 2 --exec "addprop author/string {}"

                   Hint: You may use multiple --id, --type and --owner arguments
                         and each can contain wildcards.

                   WARNING: You may very well become addicted to this command ;-)
                """   
                if not expanded :
                        return self.errormessage("Needs one Folder to find from")
                options, args = self.getopt(["id+", "maxdepth=", "owner+", "mmin=", "mtime=", "newer=", "older=", "type+", "exec="], expanded[1:])
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                def do_find(obj, zopeshell = self, options=options, args=args) :
                        if obj is not None:
                                # tests if the object matches the present options
                                many = zopeshell.match_Many(obj, options)
                                if many <= 0 :
                                        return many

                                url = zopeshell.ObjectPath(obj)
                                if options.has_key("exec") :
                                        if url is not None :
                                                return zopeshell.execCommand(string.replace(options["exec"], "{}", url)) or 0
                                        else :
                                                return zopeshell.errormessage("Error while accessing to object %s" % repr(obj))

                                if url is not None :
                                        zopeshell.htmlmessage(url)
                                        zopeshell.printf("%s\n" % url)
                                else :
                                        return zopeshell.errormessage("Error while accessing to object %s" % repr(obj))
                        return 0

                mx = self.getMaxDepth(options, default=0)
                if mx < 0 :
                        # an error occured, message already displayed
                        return -1
                return self.descend(self.toObject(self.__context, expanded[0]), do_find, maxdepth=mx)

        def run_catalog(self, expanded, unexpanded) :
                """Catalogs objects in the nearest ZCatalog

                   Accepts many arguments:

                   catalog *_dtml MyFolder/*gif [...]

                   WARNING: All objects may not end in the same
                            ZCatalog, because the choice of
                            the nearest ZCatalog is done for each
                            object.
                """
                if not expanded :
                        return self.errormessage("Needs at least one object id")
                status = 0
                for arg in expanded :
                        object = self.toObject(self.__context, arg)
                        if object is None :
                                status = status + self.errormessage('Incorrect path: %s' % arg)
                        else :
                                # finds all available Catalogs up through the acquisition path
                                zcatalog = self.__context.superValues(["ZCatalog"])
                                if not zcatalog :
                                        status = status + self.errormessage("No ZCatalog available for object %s" % self.ObjectPath(object))
                                else :
                                        zcatalog = zcatalog[0]  # we take the first one, which is the nearest
                                        if not self.HasPerms(zcatalog, 'Manage ZCatalog Entries') :
                                                status = status - 1
                                        else :
                                                zcatalog.catalog_object(object, "%s" % self.ObjectPath(object))
                                                self.htmlmessage('%s added to ZCatalog %s' % (self.ObjectPath(object), self.ObjectPath(zcatalog)))
                return status

        def run_uncatalog(self, expanded, unexpanded) :
                """Uncatalogs objects from the nearest ZCatalog

                   Accepts many arguments:

                   uncatalog *_dtml MyFolder/*gif [...]

                   Caveats: Uncatalogging an object which is not
                            present in the ZCatalog is accepted.

                   WARNING: All objects may not be uncatalogged from
                            the same ZCatalog, because the choice of
                            the nearest ZCatalog is done for each
                            object.
                """
                if not expanded :
                        return self.errormessage("Needs at least one object id")
                status = 0
                for arg in expanded :
                        object = self.toObject(self.__context, arg)
                        if object is None :
                                status = status + self.errormessage('Incorrect path: %s' % arg)
                        else :
                                # finds all available Catalogs up through the acquisition path
                                zcatalog = self.__context.superValues(["ZCatalog"])
                                if not zcatalog :
                                        status = status + self.errormessage("No ZCatalog available for object %s" % self.ObjectPath(object))
                                else :
                                        zcatalog = zcatalog[0]  # we take the first one, which is the nearest
                                        if not self.HasPerms(zcatalog, 'Manage ZCatalog Entries') :
                                                status = status - 1
                                        else :
                                                # WARNING: we don't verify if the object was really in the ZCatalog
                                                zcatalog.uncatalog_object("%s" % self.ObjectPath(object))
                                                self.htmlmessage('%s removed from ZCatalog %s' % (self.ObjectPath(object), self.ObjectPath(zcatalog)))
                return status

        def run_lsperms(self, expanded, unexpanded) :
                """Lists permissions on an object

                   Accepts an object id as its first argument.

                   Accepts the permissions to list as its other
                   arguments:

                   lsperms /QuickSart Change*

                   This will list all permissions which name matches
                   Change* on the /QuickStart object
                """   
                if not unexpanded :
                        return self.errormessage("Needs at least one object id")
                object = self.toObject(self.__context, unexpanded[0])
                if object is None :
                        return self.errormessage("Object %s doesn't exist" % unexpanded[0])
                if not self.HasPerms(object, "Change permissions") :
                        return -1
                if len(unexpanded) == 1:
                        # no permission name, we want all of them
                        unexpanded.append("*")
                result = []
                psettings = object.permission_settings()
                for perm in object.ac_inherited_permissions(all=1) :
                        pname = perm[0]
                        for pattern in unexpanded[1:] :
                                # I'd prefer to have an fnmatch.fnmatchUNcase()
                                # to be less strict...
                                if fnmatch.fnmatchcase(pname, pattern) :
                                        roles = filter(None, map(lambda r: (r["selected"] == 'SELECTED') and r["name"], object.rolesOfPermission(pname)))
                                        acquired = ((filter(lambda p, pn=pname: p["name"] == pn, psettings)[0]["acquire"] == 'CHECKED') and 'Yes') or 'No'
                                        result.append({ "Permission": pname, "Roles": string.join(roles, ', '), "Acquired": acquired })
                                        break
                self.tableDisplay("lsperms", ["Permission", "Roles", "Acquired"], result)

        def run_setperms(self, expanded, unexpanded) :
                """Sets permissions on an object

                   Accepts a --noacquire option to not acquire
                   permissions from the parent object.

                   Accepts an object id as its following argument.

                   Accepts a comma separated list of roles to which
                   give this permission. Each role not in the list
                   will have this permission removed.

                   Accepts a list of permissions as its following
                   arguments:

                   setperms --noacquire /MyForum Manager,Editor *Postings*

                   This will give all permissions which name contains Postings
                   to roles Manager and Editor on the object /MyForum
                   For other roles this permission will be removed, and will
                   not be acquired from the parent object.
                """   
                options, args = self.getopt(["noacquire"], unexpanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                if len(args) == 2 :
                        # no permission name, we want all of them
                        args.append("*")
                if len(args) < 3 :
                        return self.errormessage("Needs an object id, a comma separated list of roles and a list of permissions")
                object = self.toObject(self.__context, args[0])
                if object is None :
                        return self.errormessage("Object %s doesn't exist" % args[0])
                if not self.HasPerms(object, "Change permissions") :
                        return -1
                roles = filter(None, map(string.strip, string.split(args[1], ',')))
                acquire = not options.has_key("noacquire")
                for perm in object.ac_inherited_permissions(all=1) :
                        pname = perm[0]
                        for pattern in unexpanded[1:] :
                                # I'd prefer to have an fnmatch.fnmatchUNcase()
                                # to be less strict...
                                if fnmatch.fnmatchcase(pname, pattern) :
                                        object.manage_permission(permission_to_manage=pname, roles=roles, acquire=acquire)
                                        self.htmlmessage("Permission '%s' on object %s was given to roles %s and %s acquired otherwise" % (pname, self.ObjectPath(object), roles, (((not acquire) and 'not') or '')))
                                        break

        def run_passwd(self, expanded, unexpanded) :
                """Change or set a user's password

                   Accepts a user name as its first argument.
                   Accepts a password as its second argument.

                   If only one argument is given, then the
                   current user's password is changed:

                   passwd 67G.FDKea

                   or:

                   passwd jerome lD31

                   Caveats: Expects the user to exist in the nearest user folder.
                """   
                if len(unexpanded) == 1 :
                        username = None
                        passwd = unexpanded[0]
                elif len(unexpanded) == 2 :
                        username = unexpanded[0]
                        passwd = unexpanded[1]
                else :
                        return self.errormessage("Needs a password or an username + password")
                if not self.HasPerms(self.__context.acl_users, 'Manage users') :
                        return -1
                if username is None :
                        # no username: get the logged in user's one
                        username = AccessControl.getSecurityManager().getUser().getUserName()
                user = self.__context.acl_users.getUser(username)
                if user is None :
                        return self.errormessage("User %s doesn't exist" % username)
                roles = user.getRoles()
                domains = user.getDomains()
                self.__context.acl_users._changeUser(username, passwd, passwd, roles, domains)
                self.htmlmessage("Password changed for user %s" % username)

        def run_domains(self, expanded, unexpanded) :
                """Changes/delete allowed domains for a user

                   Accepts a user name as its first argument.
                   If no other argument is given, then allowed
                   domains for this user are deleted: the user
                   can connect from anywhere.
                   Any other argument is treated as a list of
                   domains to be ADDED to the existing list of domains
                   allowed for this user:

                   domains jerome librelogiciel.com

                   Caveats: Expects the user to exist in the nearest user folder.
                """   
                if not unexpanded :
                        return self.errormessage("Needs at least an username")
                if not self.HasPerms(self.__context.acl_users, 'Manage users') :
                        return -1
                username = unexpanded[0]
                user = self.__context.acl_users.getUser(username)
                if user is None :
                        return self.errormessage("User %s doesn't exist" % username)
                roles = list(user.getRoles())
                domains = list(user.getDomains())
                if len(unexpanded) > 1 :
                        # there are domains arguments
                        # so we add them
                        domains.extend(unexpanded[1:])
                        msg = "changed to %s" % repr(domains)
                else :
                        # there's no domain argument
                        # so we delete all domains
                        domains = []
                        msg = "deleted"

                self.__context.acl_users._changeUser(username, None, None, roles, domains)
                self.htmlmessage("Domains for user %s %s" % (username, msg))

        def run_roles(self, expanded, unexpanded) :
                """Changes/delete roles for a user

                   Accepts a user name as its first argument.
                   If no other argument is given, then roles
                   for this user are deleted: the user has no
                   role.
                   Any other argument is treated as a list of
                   roles to be ADDED to this user:

                   roles jerome Manager Editor

                   Caveats: Expects the user to exist in the nearest user folder.
                """   
                if not unexpanded :
                        return self.errormessage("Needs at least an username")
                if not self.HasPerms(self.__context.acl_users, 'Manage users') :
                        return -1
                username = unexpanded[0]
                user = self.__context.acl_users.getUser(username)
                if user is None :
                        return self.errormessage("User %s doesn't exist" % username)
                roles = list(user.getRoles())
                domains = list(user.getDomains())
                if len(unexpanded) > 1 :
                        # there are roles arguments
                        # so we add them
                        roles.extend(unexpanded[1:])
                        msg = "changed to %s" % repr(roles)
                else :
                        # there's no role argument
                        # so we delete all roles
                        roles = []
                        msg = "deleted"
                self.__context.acl_users._changeUser(username, None, None, roles, domains)
                self.htmlmessage("Roles for user %s %s" % (username, msg))

        def run_call(self, expanded, unexpanded) :
                """Calls an object with optional parameters

                   The object name may be specified with its full path with dots or slashes:

                   call MyFolder.MyObject(_,_.REQUEST,arg1="thisarg"[,arg2="anotherarg"][,...])

                   or:

                   call MyFolder.MyObject(context,context.REQUEST,arg1="thisarg"[,arg2="anotherarg"][,...])

                   another example:

                   call index_html

                   This last one will return the unrendered version of index_html

                   Both _ and context may be used if needed.

                   Hint: Don't use any space in your command, or else enclose your
                         command between single or double quotes.

                   WARNING: Calling zshell itself is a Very Bad Idea (tm) and can kill Zope
                """   
                if not expanded :
                        return self.errormessage("Needs an object id to call")

                # Michel@DC: you should factor the object out of this eval and
                # validate it with
                # SecurityManager.checkPermission('View', object).
                # Also, 'eval' without an namespace qualifying 'in'
                # clause can be bad!  Try and do this without eval.

                # Jerome: Don't know how without eval !
                # new code looks very ugly and accessing to object's
                # properties doesn't work anymore, unfortunately.

                objectstr = string.join(unexpanded, ' ')
                pos = string.find(objectstr, '(')
                if pos == -1 :
                        # called without arguments
                        objpath = objectstr
                        objargs = ""
                else :
                        # called with arguments, skip them
                        # because we only want the object name
                        objpath = objectstr[:pos]
                        objargs = objectstr[pos:]

                objpath = string.replace(objpath, '.', '/')
                object = self.toObject(self.__context, objpath)
                if object is None :
                        # maybe should do something to re-allow properties to be used
                        return self.errormessage("Object %s not found" % objectstr)
                else :
                        if not self.HasPerms(object, 'View') :
                                return -1
                        else :
                                _ = context = self.__context
                                callresult = str(eval("object%s" % objargs))
                                self.printf("%s" % callresult)

        def run_lsusers(self, expanded, unexpanded) :
                """Lists users in the nearest User Folder

                   Accepts multiple user names as its arguments,
                   wildcards are accepted:

                   lsusers user1 [jer*] [...]
                """
                if not unexpanded :
                        unexpanded = ["*"]      # List all users
                if not self.HasPerms(self.__context.acl_users, 'Manage users') :
                        return -1
                result = []
                for username in self.__context.acl_users.getUserNames() :
                        for uname in unexpanded :
                                if fnmatch.fnmatchcase(username, uname) :
                                        user = self.__context.acl_users.getUser(username)
                                        result.append({ "UserName": username, "Roles": string.join(user.getRoles(), ', '), "InContext": string.join(user.getRolesInContext(self.__context), ', '), "Domains": string.join(user.getDomains(), ', ') })
                self.tableDisplay("lsusers", ["UserName", "Roles", "InContext", "Domains"], result)

        def run_delusers(self, expanded, unexpanded):
                """Delete users from the nearest User Folder

                   Accepts multiple user names as its arguments:

                   delusers user1 [user2] [...]

                   WARNING: No wildcard is expanded, for security reasons.
                """
                if not expanded :
                        return self.errormessage('Needs an userid as the first argument')
                if not self.HasPerms(self.__context.acl_users, 'Manage users') :
                        return -1
                status = 0
                usernames = []
                for username in unexpanded :
                        if username not in self.__context.acl_users.getUserNames() :
                                status = status + self.errormessage("User %s doesn't exists" % username)
                        else :
                                usernames.append(username)

                self.__context.REQUEST.set("names", usernames)
                self.__context.acl_users.manage_users("Delete", REQUEST=self.__context.REQUEST)
                if usernames :
                        self.htmlmessage('Users %s deleted' % string.join(usernames, ", "))

                # don't be fucked by Zope's automatic redirection
                self.__context.REQUEST.RESPONSE.setStatus(200)
                return status

        def run_addusers(self, expanded, unexpanded) :
                """Adds users into the nearest User Folder

                   The users added have no role, and no domain.
                   Accepts multiple arguments, each argument must be
                   made of a username, a slash character, and a password:

                   addusers jerome/kyx.ud34 [john/9!AZce] [...]
                """
                if not unexpanded :
                        return self.errormessage('Needs an userid/password as the first argument')
                if not self.HasPerms(self.__context.acl_users, 'Manage users') :
                        return -1
                status = 0
                for arg in unexpanded :
                        split = string.split(arg, '/')
                        if len(split) != 2 :
                                status = status + self.errormessage('Incorrect username/password: %s' % arg)
                        else :
                                (username, password) = split
                                if username in self.__context.acl_users.getUserNames() :
                                        status = status + self.errormessage("User %s already exists" % username)
                                else :
                                        self.__context.REQUEST.set("name", username)
                                        self.__context.REQUEST.set("password", password)
                                        self.__context.REQUEST.set("confirm", password)
                                        self.__context.REQUEST.set("domains", "")
                                        self.__context.REQUEST.set("roles", [])
                                        self.__context.acl_users.manage_users("Add", REQUEST=self.__context.REQUEST)
                                        self.htmlmessage('User %s added with password %s' % (username, password))
                # don't be fucked by Zope's automatic redirection
                self.__context.REQUEST.RESPONSE.setStatus(200)
                return status

        def run_lroles(self, expanded, unexpanded) :
                """Sets local roles for an user in the local Folder

                   Accepts the user's name as its first argument
                   Accepts a list of local roles to give to this user
                   as its remaining arguments, if no role name is given,
                   then all local roles for this user are deleted:

                   lroles jerome Manager [Moderator] [...]
                """
                if not unexpanded :
                        return self.errormessage('Needs an userid as the first argument')
                username = unexpanded[0]
                if len(unexpanded) > 1 :
                        roles = unexpanded[1:]
                        # Zope accepts even if user and roles don't exist at all
                        # so we have to test it ourselves
                        if username not in self.__context.acl_users.getUserNames() :
                                return self.errormessage("Unknown user %s" % username)
                        if not self.HasPerms(self.__context, "Change permissions") :
                                return -1
                        # should we also test if roles exits ?
                        self.__context.manage_setLocalRoles(userid=username, roles=roles)
                        self.htmlmessage('User %s now has local roles: %s' % (username, string.join(roles, ', ')))
                else :
                        self.__context.manage_delLocalRoles(userids=[username])
                        self.htmlmessage('User %s now has no local role' % username)

        def run_whoami(self, expanded, unexpanded) :
                """Shows the current username"""
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                (username, roles) = self.WhoAmI()
                self.printf('Username: %s &nbsp;&nbsp;&nbsp;&nbsp; Roles: %s' % (username, string.join(roles, ', ')))

        def run_pack(self, expanded, unexpanded) :
                """Packs the ZODB

                   Accepts a number of days as its single argument:

                   pack 3
                """
                if not self.HasRoles(self.__context.Control_Panel, 'Manager') :
                        return -1
                if len(unexpanded) > 1 :
                        return self.errormessage('Needs a number of days as an argument')
                elif not unexpanded :
                        ndays = 0
                else :
                        ndays = int(unexpanded[0])
                try:
                        self.__context.Control_Panel.manage_pack(days = ndays)
                except FileStorage.FileStorageError :
                        pass    # no pack needed, but we don't want the error message
                self.htmlmessage("Database packed")

        def run_restart(self, expanded, unexpanded) :
                """Restarts Zope

                   Works fine but unfortunately Zope redirects
                   us automatically to zshell/manage_main
                """
                if not self.HasRoles(self.__context.Control_Panel, 'Manager') :
                        return -1
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                self.__context.Control_Panel.manage_restart(self.__context.REQUEST.URL0)

        def run_shutdown(self, expanded, unexpanded) :
                """Shutdowns Zope

                   Works fine but unfortunately looks ugly.
                """
                if not self.HasRoles(self.__context.Control_Panel, 'Manager') :
                        return -1
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                self.__context.Control_Panel.manage_shutdown()

        def run_davlocks(self, expanded, unexpanded) :
                """Returns Zope's DataBase WebDAV locks

                   accepts paths to Folder objects as its arguments.

                   all paths given will be searched recursively to find WebDAV locks.

                   if no path is given then the search will begin in the current Folder.
                """
                if not self.HasRoles(self.__context.Control_Panel, 'Manager') :
                        return -1
                if not expanded :
                        expanded.append(self.ObjectPath(self.__context))
                status = 0
                lockscount = 0
                for path in expanded :
                        object = self.toObject(self.__context, path)
                        if object is None :
                                status = status + self.errormessage("Object %s doesn't exist" % path)
                        else :
                                locked = self.__context.Control_Panel.DavLocks.findLockedObjects(self.ObjectPath(object))
                                lockscount = lockscount + len(locked)
                                for (url, locklist) in locked :
                                        for l in locklist :
                                                # Looks ugly : self.htmlmessage("%s locked by %s (token=%s)" % (url, l["owner"], l["token"]), printable=1)
                                                self.htmlmessage("%s locked by %s" % (url, l["owner"]), printable=1)
                if not lockscount :
                        self.htmlmessage("No WebDAV lock found !", printable=1)
                else :
                        self.htmlmessage("%i WebDAV locks found." % lockscount, printable=1)
                return status

        def run_dbconn(self, expanded, unexpanded) :
                """Returns Zope's DataBase active connections.
                """
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                conns = self.__context.Control_Panel.DebugInfo.dbconnections()
                for conn in conns :
                        self.htmlmessage(str(conn), printable=1)

        def run_tid(self, expanded, unexpanded) :
                """Returns Zope's current thread id.
                """
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                self.htmlmessage(str(self.__context.Control_Panel.thread_get_ident()), printable=1)

        def run_dbname(self, expanded, unexpanded) :
                """Returns Zope's DataBase name.
                """
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                self.htmlmessage(self.__context.Control_Panel.db_name(), printable=1)

        def run_dbsize(self, expanded, unexpanded) :
                """Returns Zope's DataBase size.
                """
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                self.htmlmessage(self.__context.Control_Panel.db_size(), printable=1)

        def run_uptime(self, expanded, unexpanded) :
                """Returns Zope's uptime.
                """
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                self.htmlmessage(self.__context.Control_Panel.process_time(), printable=1)

        def run_echo(self, expanded, unexpanded) :
                """Prints its argument followed by a newline.

                   Accepts multiple arguments.

                   e.g. :

                     echo Hello ZopeWorld !
                """
                self.printf(string.join(expanded, " ") + "\n")

        def run_printenv(self, expanded, unexpanded) :
                """Prints environment variables.

                   Accepts multiple arguments.

                   printenv [var] [...]

                   e.g. :

                     printenv PATH
                """
                all = 0
                if not expanded :
                        expanded = self.__env.keys()
                        all = 1
                for arg in expanded :
                        value = self.__env.get(arg)
                        if value is not None :
                                if all :
                                        self.htmlmessage("%s=%s" % (arg, str(value)), printable=1)
                                else :
                                        self.htmlmessage("%s" % str(value), printable=1)

        def run_unsetenv(self, expanded, unexpanded) :
                """Deletes variables from the environment.

                   Accepts multiple arguments.

                   unsetenv [--zope] var [...]

                     --zope : changes Zope's running environment too,
                              which also render variables persistent from
                              one call to ZShell to another.

                   e.g. :

                     unsetenv PATH

                   WARNING : the --zope option's possible collateral damage is currently
                             under investigation... Please report any problem.
                """
                options, args = self.getopt(["zope"], expanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                if not args :
                        return self.errormessage("Needs at least one variable to unset")
                for arg in args :
                        # no need to be verbose when the var doesn't exist
                        # because in a sense the command is already successful ;-)
                        if self.__env.has_key(arg) :
                                del self.__env[arg]
                        if options.has_key("zope") and os.environ.has_key(arg) :
                                del os.environ[arg]

        def run_setenv(self, expanded, unexpanded) :
                """Sets envronment variables.

                   Accepts multiple arguments.

                   setenv [--zope] var1=value1 [var2=value2]

                     --zope : changes Zope's running environment too,
                              which also render variables persistent from
                              one call to ZShell to another.

                   e.g. :

                     setenv PATH=/var/zope

                   WARNING : the --zope option's possible collateral damage is currently
                             under investigation... Please report any problem.
                """
                options, args = self.getopt(["zope"], expanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                if not args :
                        return self.errormessage("Needs at least one variable to set")
                status = 0
                for arg in args :
                        var = string.split(arg, '=')
                        if len(var) == 2 :
                                (name, value) = var
                                value = self.expandvar(value)
                                if options.has_key("zope") :
                                        os.environ[name] = value
                                self.__env[name] = value
                        else :
                                status = status + self.errormessage("Invalid syntax '%s'" % arg)
                return status

        def run_mkver(self, expanded, unexpanded) :
                """Create versions

                   Accepts multiple arguments:

                   mkver debug [stable] [...]
                """
                if not unexpanded :
                        return self.errormessage('Needs a version id as the first argument')
                if not self.HasPerms(self.__context, 'Add Versions') :
                        return -1
                status = 0
                for vid in unexpanded :
                        if vid in self.__context.objectIds() :
                                status = status + self.errormessage('Object %s already exists' % vid)
                        else :
                                self.__context.manage_addProduct["OFSP"].manage_addVersion(id = vid, title = vid)
                                self.htmlmessage('Version %s created' % vid)
                return status

        def run_enter(self, expanded, unexpanded) :
                """Enter into a version

                   Accepts a version id (with an optional path) as its only argument

                   enter debug

                   Caveats: it seems you're not really in the version until the end of the
                            transaction, so commands entered after 'enter' will work outside
                            of this version.
                """
                if not expanded :
                        return self.errormessage('Needs a version id as an argument')
                vexist = self.toObject(self.__context, expanded[0])
                if not vexist :
                        return self.errormessage("Version <em><b>%s</b></em> doesn't exist" % expanded[0])
                else :
                        if not self.HasPerms(vexist, 'Join/leave Versions') :
                                return -1
                        vexist.enter(self.__context.REQUEST, self.__context.REQUEST.RESPONSE)

                        # get_transaction().commit(1) doesn't seem to do it !

                        # don't be fucked by Zope's automatic redirection
                        self.__context.REQUEST.RESPONSE.setStatus(200)
                        self.htmlmessage("You'll be working in version %s at the end of the current transaction" % self.ObjectPath(vexist))

        def run_leave(self, expanded, unexpanded) :
                """Quits a version

                   Accepts a version id as its only argument

                   leave debug

                   Caveats: it seems you're not really outside of the version until the end of the
                            transaction, so commands entered after 'leave' will work inside
                            this version.
                """
                if not expanded :
                        return self.errormessage('Needs a version id as an argument')
                vexist = self.toObject(self.__context, expanded[0])
                if not vexist :
                        return self.errormessage("Version <em><b>%s</b></em> doesn't exist" % expanded[0])
                else :
                        if not self.HasPerms(vexist, 'Join/leave Versions') :
                                return -1
                        vexist.leave(self.__context.REQUEST, self.__context.REQUEST.RESPONSE)

                        # get_transaction().commit(1) doesn't seem to do it !

                        # don't be fucked by Zope's automatic redirection
                        self.__context.REQUEST.RESPONSE.setStatus(200)
                        self.htmlmessage("You'll not be working in version %s anymore at the end of the current transaction" % self.ObjectPath(vexist))

        def run_save(self, expanded, unexpanded) :
                """Commits a version's changes

                   Accepts a commit message as its optional argument:

                   save The new version seems to be OK

                   WARNING: Needs testers
                """
                if not self.__version :
                        return self.errormessage("Not in a version")
                else :
                        objver = self.toObject(self.__context, self.__version)
                        if objver is None :
                                return self.errormessage("Error while accessing version %s" % self.__version)
                        else :
                                if not self.HasPerms(objver, 'Save/discard Version changes') :
                                        return -1
                                # for save, remark doesn't have a default value (according to Zope 2.3.0 sources)
                                objver.save(remark = (string.join(expanded, ' ') or 'No comment'))
                                self.htmlmessage("Version %s saved" % self.ObjectPath(objver))

        def run_discard(self, expanded, unexpanded) :
                """Discards a version's changes

                   Accepts a discard message as its optional argument:

                   discard I was doing completely wrong

                   WARNING: Needs testers
                """
                if not self.__version :
                        return self.errormessage("Not in a version")
                else :
                        objver = self.toObject(self.__context, self.__version)
                        if objver is None :
                                return self.errormessage("Error while accessing version %s" % self.__version)
                        else :
                                if not self.HasPerms(objver, 'Save/discard Version changes') :
                                        return -1
                                # for discard, remark's default value is an empty string
                                objver.discard(remark = string.join(expanded, ' '))
                                self.htmlmessage("Version %s discarded" % self.ObjectPath(objver))

        def run_copy(self, expanded, unexpanded) :
                """Copy objects to the clipboard

                   Accepts multiple objects ids as its arguments, but
                   each object must be in the current Folder:

                   copy obj1 [obj2] [...]
                """
                if not expanded :
                        return self.errormessage("Needs some objects ids to copy")
                if not self.HasPerms(self.__context, 'View management screens') :
                        return -1
                status = 0
                objids = []
                for objid in expanded :
                        if '/' in objid :
                                status = status + self.errormessage('Paths for objects ids are not allowed at this time: %s' % objid)
                        else :
                                objids.append(objid)
                try :
                        self._clipboard = self.__context.manage_copyObjects(ids = objids)
                        for objid in objids :
                                self.htmlmessage('%s copied to clipboard' % objid)
                except AttributeError, msg :
                        status = status + self.errormessage("Object %s doesn't exist" % msg)
                return status

        def run_cut(self, expanded, unexpanded) :
                """Cut objects to the clipboard

                   Accepts multiple objects ids as its arguments, but
                   each object must be in the current Folder:

                   cut obj1 [obj2] [...]
                """
                if not expanded :
                        return self.errormessage("Needs some objects ids to cut")
                if not self.HasPerms(self.__context, 'View management screens') :
                        return -1
                status = 0
                objids = []
                for objid in expanded :
                        if '/' in objid :
                                status = status + self.errormessage('Paths for objects ids are not allowed at this time: %s' % objid)
                        else :
                                objids.append(objid)
                try :
                        self._clipboard = self.__context.manage_cutObjects(ids = objids)
                        for objid in objids :
                                self.htmlmessage('%s cut to clipboard' % objid)
                except AttributeError, msg :
                        status = status + self.errormessage("Object %s doesn't exist" % msg)
                return status

        def run_paste(self, expanded, unexpanded) :
                """Paste the clipboard's contents into the current Folder"""
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                if not self.HasPerms(self.__context, 'View management screens') :
                        return -1
                if not hasattr(self, '_clipboard') :
                        return self.errormessage("Clipboard is empty")
                try :
                        self.__context.manage_pasteObjects(cb_copy_data = self._clipboard)
                        self.htmlmessage("Clipboard's content pasted into %s" % self.getcwd())
                except CopyError :
                        return self.errormessage("Impossible to paste clipboard's content into %s" % self.getcwd())

        def run_dump(self, expanded, unexpanded) :
                """Exports objects source to a directory on the server

                   Accepts a destination directory as its first
                   argument. The destination directory must exist
                   on the server.

                   Accepts multiple objects ids to export as its
                   following arguments.

                   dump /home/jerome/dtml /*_html /*_dtml

                   Hint:    Be careful to give sufficient permissions to
                            the user which Zope runs as on the destination
                            directory.

                   Caveats: Only objects which have a callable
                            document_src attribute can be exported this way.
                """         
                if len(expanded) < 2 :
                        return self.errormessage("Needs at least a destination directory and one object id to dump")
                destination = os.path.normpath(os.path.expanduser(expanded[0])) # in case there's a ~username
                if not os.path.isdir(destination) :
                        return self.errormessage("%s is not a directory" % destination)
                status = 0
                for arg in expanded[1:] :
                        object = self.toObject(self.__context, arg)
                        if object is None :
                                status = status + self.errormessage("Object %s doesn't exist" % arg)
                        elif not self.HasPerms(object, 'View management screens') :
                                status = status - 1
                        elif not hasattr(object, "document_src") or not callable(object.document_src) :
                                status = status + self.errormessage("Doesn't know how to dump object %s" % arg)
                        else :
                                fname = os.path.join(destination, object.getId())
                                try :
                                        fout = open(fname, "wb")
                                        fout.write(object.document_src())
                                        fout.close()
                                        self.htmlmessage("Object %s dumped to server as %s" % (self.ObjectPath(object), fname))
                                except IOError, msg :
                                        status = status + self.errormessage('Error %s, occured while dumping %s' % (msg, arg))
                return status

        def run_export(self, expanded, unexpanded) :
                """Exports objects to the server

                   Accepts a --xml option to export as xml
                   data.

                   Accepts multiple objects ids to export

                   export --xml MyPath/Myobject /QuickStart

                   Caveats: Exports exclusively to the server.
                            The Root Folder is exported as .zexp or .xml
                """         
                options, args = self.getopt(["xml"], expanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                if not args :
                        return self.errormessage("Needs at least one object id to export")
                status = 0
                for arg in args :
                        object = self.toObject(self.__context, arg)
                        if object is None :
                                status = status + self.errormessage("Object %s doesn't exist" % arg)
                        elif not hasattr(object, "aq_parent") :
                                status = status + self.errormessage("Object %s is not exportable" % arg)
                        elif not self.HasPerms(object.aq_parent, 'Import/Export objects') :
                                status = status - 1
                        else :
                                toxml = 0
                                download = 0    # TODO: Zope 2.3.2 is buggy, so don't allow downloads yet
                                if options.has_key("xml") :
                                        toxml = 1
                                object.aq_parent.manage_exportObject(id=object.getId(), download=download, toxml=toxml)
                                fname = "%s.%s" % (object.getId(), (toxml and 'xml') or 'zexp')
                                self.htmlmessage("Object %s exported to server as %s" % (self.ObjectPath(object), fname))
                return status

        def run_import(self, expanded, unexpanded) :
                """Imports objects into the current Folder

                   Accepts multiple filenames to import, even
                   with a leading path. filename resolution is
                   made server side.

                   import one.zexp [path/to/two.zexp] [...]

                   Warning : filename resolution will be done in the
                   following order :
                     1 - as-is
                     2 - from $INSTANCE_HOME/import
                     3 - from $SOFTWARE_HOME/../../import
                """   
                if not unexpanded :
                        return self.errormessage("Needs some filenames to import")
                if not self.HasPerms(self.__context, 'Import/Export objects') :
                        return -1
                instance_home_import = os.path.join(INSTANCE_HOME, '%simport' % os.sep)
                instance_home_import = os.path.normpath(instance_home_import)
                software_home_import = os.path.join(SOFTWARE_HOME, '..%s..%simport' % (os.sep, os.sep))
                software_home_import = os.path.normpath(software_home_import)
                for arg in unexpanded :
                        (dirname, fname) = os.path.split(arg)
                        for impath in (dirname, instance_home_import, software_home_import) :
                                filepath = os.path.join(impath, fname)
                                if os.path.exists(filepath) :
                                        globpath = [ filepath ]
                                        break
                                else :
                                        globpath = glob.glob(filepath)
                                        if not globpath :
                                                globpath = [ filepath ]
                                        else :
                                                break
                        for filename in globpath :
                                if os.path.exists(filename) :
                                        self.__context._importObjectFromFile(filename)
                                        self.htmlmessage('%s imported successfully' % filename)

        def run_takeown(self, expanded, unexpanded) :
                """Take ownership

                   Optionally accepts --recurse as an option to ask for a recursive action.
                   Accepts multiple pathnames as its other arguments:

                   takeown [--recurse] path/to/obj1 [otherpath/obj2] [...]

                   Caveats: Due to a bug in Zope 2.3.2, taking ownership
                            recursively from a folderish object you already own
                            does nothing, even if you don't already own the full
                            subtree. You should try to do it with the find command
                            instead if you don't want Zope's default buggy behavior.
                """
                options, args = self.getopt(["recurse"], expanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                recursive = 0
                recursive_msg = ""
                if options.has_key("recurse") :
                        recursive = 1
                        recursive_msg = "recursively "
                if not args :
                        return self.errormessage("Needs at least one object id")
                chownto = AccessControl.getSecurityManager().getUser()
                status = 0
                for objpath in args :
                        object = self.toObject(self.__context, objpath)
                        if object is None :
                                status = status + self.errormessage('Incorrect path: %s' % objpath)
                        elif not self.HasPerms(object, 'Take ownership') :
                                status = status - 1
                        else :
                                object.changeOwnership(chownto, recursive = recursive)
                                self.htmlmessage('%s owner %schanged to %s' % (self.ObjectPath(object), recursive_msg, chownto.getUserName()))
                return status

        def run_man(self, expanded, unexpanded) :
                """Displays command's man pages

                   Accepts several commands' names as its arguments
                   or none to list all commands:

                   man [cmd1] [...]
                """   
                methodslist = map(lambda n: n[4:], filter(lambda f: f[:4] == 'run_', self.__class__.__dict__.keys()))
                if not unexpanded :
                        unexpanded = methodslist
                unexpanded.sort()

                results = []
                for method in unexpanded :
                        if results :
                                # more than one command to display help for
                                # so we separate them in the plain text output
                                self.printf("\n--------\n\n")
                        if not hasattr(self, 'run_' + method) :
                                help = helphtml = "Invalid command"
                        else :
                                help = getattr(self, 'run_' + method).__doc__
                                if not help :
                                        help = helphtml = "Undocumented command"
                                else :
                                        helplines = map(string.strip, string.split(help, '\n'))
                                        help = string.join(helplines, '\n')
                                        helphtml = string.join(helplines, '<br />')
                        command = '<a href="%s?zshellscript=man%%20%s&zshellscript=%s&zshelldontrun=1">%s</a>' % (self.__context.REQUEST.URL0, method, method, method)
                        results.append({"Command": command, "Help": helphtml})
                        self.printf("%s: %s\n" % (method, help))
                self.tableDisplay("man", ["Command", "Help"], results)

        def run_whatis(self, expanded, unexpanded) :
                """An alias to man"""
                return self.run_man(expanded, unexpanded)

        def run_help(self, expanded, unexpanded) :
                """An alias to man"""
                return self.run_man(expanded, unexpanded)

        def run_apropos(self, expanded, unexpanded) :
                """An alias to man"""
                return self.run_man(expanded, unexpanded)

        def run_about(self, expanded, unexpanded) :
                """About this software"""
                msg =  "ZShell v%s" % __version__
                url = "http://www.librelogiciel.com/software/"
                who = "Jerome Alet"
                email = "alet@librelogiciel.com"
                self.printf("%s (%s) (C) 2001 %s - %s\n%s\n" % (msg, url, who, email, __doc__))

        def run_zhelp(self, expanded, unexpanded) :
                """Search terms in Zope's internal help

                   Accepts multiple arguments which are concatenated together:

                   zhelp ZClass
                """
                self.__context.REQUEST.set("SearchableText", string.join(unexpanded, " "))
                results = self.__context.HelpSys.results(self.__context, self.__context.REQUEST)
                # we just want the result lines begining with an <a href=" tag, the rest is uninteresting
                self.htmlmessage(string.join(filter(lambda r: r[:9] == '<a href="', string.split(results, "\n")), "\n"), safe=1)

        def run_wget(self, expanded, unexpanded) :
                """Sucks documents from the web and import them in the current Folder

                   Accepts multiple arguments, and can retrieve files from the
                   filesystem, or even complete directories.

                   wget http://www.zope.org/ http://www.gnu.org/graphics/gnu-head-sm.jpg

                   or:

                   wget ~jerome/mydirectory/*.html /home/*/public_html http://localhost

                   Caveats: Maybe needs a recursive option, but I'm not sure:
                            it's already very powerful.
                """
                if not unexpanded:
                        return self.errormessage('Needs at least one argument')
                if not self.HasPerms(self.__context, 'Add Documents, Images, and Files') :
                        return -1

                # expand arguments from the filesystem, not the ZODB
                # in the case there's a file: scheme or no scheme at all.
                expanded = []
                for uarg in unexpanded :
                        if uarg[:5] == 'file:' :
                                uarg = uarg[5:]
                        uarg = os.path.expanduser(uarg) # in case of a ~username
                        if os.path.isdir(uarg) : # it's a filesystem directory, so we want all its files
                                uarg = os.path.join(uarg, '*')
                        expanded.extend(self.ShellExpand(uarg, zodb=0) or [uarg])

                status = 0
                for arg in expanded :
                        try :
                                # WARNING: both urlopen and object.read may
                                # raise an IOError, in the latter case that's when
                                # object is a filesystem directory
                                object = urllib.urlopen(arg)
                                info = object.info()
                                ctype = info.gettype()
                                mtype = info.getmaintype()
                                stype = info.getsubtype()
                                data = object.read()
                                realurl = object.geturl()
                                if realurl[-1] == '/' :
                                        fname = 'index' + '_' + stype
                                else :
                                        fname = filter(None, string.split(realurl, '/'))[-1]
                                if fname in self.__context.objectIds() :
                                        status = status + self.errormessage('Object %s already exists' % fname)
                                else :
                                        if mtype == "image" :
                                                # Image
                                                self.__context.manage_addImage(id = fname, file = data, title = realurl, precondition = '', content_type = ctype)
                                        elif ctype == 'text/html' :
                                                # DTML Document
                                                self.__context.manage_addDTMLDocument(id = fname, title = realurl, file = data)
                                        else :
                                                # normal File
                                                self.__context.manage_addFile(id = fname, file = data, title = realurl, precondition = '', content_type = ctype)
                                        self.htmlmessage('%s added as %s, size is %ld bytes' % (realurl, fname, len(data)))
                                del data
                                object.close()
                                del object
                        except IOError,msg :
                                status = status + self.errormessage('Error %s, occured while retrieving %s' % (msg, arg))
                return status

        def run_mkuf(self, expanded, unexpanded) :
                """Creates User Folders

                   Accepts multiple folders as its arguments:

                   mkuf /Folder1/Folder2 [...]

                   Remark: The User Folder is only created in the deepest
                           folderish object, e.g. in the example above
                           the User Folder will be created in Folder1/Folder2
                           but not in Folder1.
                """   
                if not expanded :
                        return self.errormessage("Needs at least a folder id")

                status = 0
                for arg in expanded :
                        object = self.toObject(self.__context, arg)
                        if object is not None :
                                if not object.isPrincipiaFolderish :
                                        status = status + self.errormessage("%s is not a Folderish object" % self.ObjectPath(object))
                                elif 'acl_users' in object.objectIds() :
                                        status = status + self.errormessage("%s already contains an User Folder or an object named acl_users" % self.ObjectPath(object))
                                elif not self.HasPerms(object, 'Add User Folders') :
                                        status = status - 1
                                else :
                                        object.manage_addUserFolder()
                                        self.htmlmessage("User Folder added in object %s" % self.ObjectPath(object))
                return status

        def run_mkdir(self, expanded, unexpanded) :
                """Create Folders

                   Accepts multiple pathnames as its arguments,
                   and creates Folders recursively:

                   mkdir path/to/folder/to/create [...]
                """
                if not expanded:
                        return self.errormessage('Needs at least one argument')
                status = 0
                for foldername in expanded :
                        exist = self.toObject(self.__context, foldername)
                        if exist is not None :
                                status = status + self.errormessage('Folder %s already exists' % self.ObjectPath(exist))
                                continue

                        if foldername[0] == '/' :
                                c = '/'
                                foldername = foldername[1:]
                        else :
                                c = ''
                        components = filter(None, string.split(foldername, '/'))
                        curpath = c[:]
                        oldcurpath = curpath[:]
                        error = 0
                        for component in components :
                                if curpath != c :
                                        curpath = curpath + '/' + component
                                else :
                                        curpath = curpath + component
                                if self.toObject(self.__context, curpath) is None :
                                        parent = self.toObject(self.__context, oldcurpath)
                                        if parent is None :
                                                status = status + self.errormessage('Unknown error on %s' % foldername)
                                                error = 1
                                                break
                                        if component not in parent.objectIds() :
                                                if not self.HasPerms(parent, "Add Folders") :
                                                        status = status - 1
                                                        error = 1
                                                        break
                                                else :
                                                        parent.manage_addFolder(id = component)
                                oldcurpath = curpath[:]
                        if not error :
                                self.htmlmessage('Folder %s created' % foldername)
                return status

        def run_pwd(self, expanded, unexpanded) :
                """Returns the current working Folder"""
                if expanded :
                        return self.errormessage("Doesn't need any argument")
                self.printf("Current folder is: %s\n" % self.getcwd())

        def run_cd(self, expanded, unexpanded) :
                """Change Directory

                   Accepts a full pathname as its single argument:

                   cd SubFolder/SubSubFolder
                """ 
                if len(expanded) != 1 :
                        return self.errormessage('Needs one and only one argument')
                newdir = self.toObject(self.__context, expanded[0])
                if (newdir is not None) and newdir.isPrincipiaFolderish :
                        self.__context = newdir
                        self.htmlmessage("Current folder is: %s" % self.ObjectPath(newdir))
                else :
                        self.errormessage("Incorrect path: %s" % expanded[0])
                        self.run_pwd([], [])
                        return -1

        def run_rm(self, expanded, unexpanded) :
                """Deletes objects

                   Accepts multiple pathnames as its arguments:

                   rm path1/object1 [...]

                   WARNING: deletion is recursive and you can even delete the
                   current Folder. USE AT YOUR OWN RISK !
                """
                if not expanded :
                        return self.errormessage('Needs at least one argument')
                status = 0
                for objpath in expanded :
                        object = self.toObject(self.__context, objpath)
                        if object is None :
                                status = status + self.errormessage("Incorrect path: %s" % objpath)
                        else :
                                objurl = self.ObjectPath(object)
                                if hasattr(object, 'aq_parent') :
                                        if not self.HasPerms(object.aq_parent, 'Delete objects') :
                                                status = status - 1
                                        else :
                                                if not hasattr(object.aq_parent, "manage_delObjects") or not callable(object.aq_parent.manage_delObjects) :
                                                        status = status + self.errormessage("manage_delObjects: operation not supported on %s" % self.ObjectPath(object.aq_parent))
                                                else :
                                                        object.aq_parent.manage_delObjects(ids = [object.getId()])
                                                        self.htmlmessage("%s removed" % objurl)
                                else :
                                        status = status + self.errormessage("Unknown error on: %s" % objurl)
                return status

        def run_mv(self, expanded, unexpanded) :
                """Moves objects

                   Accepts multiple source arguments
                   Each source argument must be an object id and
                   each object should be in the current Folder.
                   The last argument is the destination Folder,
                   and may be a slash delimited pathname:

                   mv obj1 [obj2] [...] destination

                   Caveats: can't rename objects yet
                """
                return self.mv_or_cp("mv", expanded)

        def run_cp(self, expanded, unexpanded) :
                """Copy objects 

                   Accepts multiple source arguments
                   Each source argument must be an object id and
                   each object should be in the current Folder.
                   The last argument is the destination Folder,
                   and may be a slash delimited pathname:

                   cp obj1 [obj2] [...] destination

                   Caveats: destination must be a Folder
                """
                return self.mv_or_cp("cp", expanded)

        def run_ls(self, expanded, unexpanded) :
                """List objects

                   Accepts multiple arguments:

                   ls Contr?l_Pan*/Products/Sq* MyFolder/*
                """
                if not expanded :
                        expanded = self.ShellExpand("*") # no argument: list all

                if expanded :
                        status = 0
                        results = []
                        for arg in expanded :
                                object = self.toObject(self.__context, arg)
                                if object is not None :
                                        if hasattr(object, 'aq_parent') :
                                                if not self.HasPerms(object.aq_parent, 'Access contents information') :
                                                        status = status - 1
                                                        continue

                                        objurl = self.ObjectPath(object)
                                        urls = '<a href="%s/manage" target="top">Manage</a>/<a href="%s" target="top">View</a>' % (objurl, objurl)
                                        ownerinfo = object.owner_info()
                                        if (ownerinfo is not None) :
                                                if hasattr(ownerinfo, "has_key") and ownerinfo.has_key('id') :
                                                        owner = ownerinfo['id']
                                                else :
                                                        # at least for /Control_Panel/Products[/*]
                                                        owner = repr(ownerinfo)
                                        else :
                                                owner = 'Not owned'

                                        try :
                                                modtime = object.bobobase_modification_time().strftime("%Y-%m-%d %H:%M:%S %Z")
                                        except :
                                                modtime = "Unknown"

                                        results.append({"Id": object.getId(), "Title": object.title, "MetaType": self.getMetaType(object), "Mod. Time": modtime, "Owner": owner, "SubObj": len(object.objectValues()), "Actions": urls })
                                        self.printf("%s\n" % objurl)
                        self.tableDisplay("ls", ["Id", "Title", "MetaType", "Mod. Time", "Owner", "SubObj", "Actions"], results)
                        return status
                else :
                        self.htmlmessage("Empty")

        def run_setprop(self, expanded, unexpanded) :
                """Sets an object's property value

                   Accepts mandatory --name and --value options, each with an argument,
                   and multiple objects ids as its remaining arguments:

                   setprop --name author --value "William Shakespeare" Hamlet_dtml Othello_?tml

                   This sets the property 'author' value to 'William Shakespeare'
                   to the objects Hamlet_dtml and Othello_?tml in the
                   current Folder.
                """
                options, args = self.getopt(["name=", "value="], expanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                if not args :
                        return self.errormessage("Needs at least one object to change this property")
                if not (options.has_key("name") and options.has_key("value")) :
                        return self.errormessage("You must supply a property name and value")
                property = options["name"]
                if options.has_key("value") :
                        propvalue = options["value"]
                        try :
                                # maybe it's a list in a string, e.g. "['e', 'f']"
                                # or something like that
                                newvalue = eval(propvalue)
                                if (type(newvalue) != type(0)) and (type(newvalue) != type(0.0)) :
                                        # we mustn't convert numeric to string
                                        propvalue = newvalue
                        except NameError :
                                pass    # normal string
                        except SyntaxError :
                                pass    # string containing spaces
                else :
                        propvalue = ""

                status = 0
                for arg in args :
                        object = self.toObject(self.__context, arg)
                        if object is not None :
                                if not self.HasPerms(object, 'Manage properties') :
                                        status = status - 1
                                elif hasattr(object, 'hasProperty') :
                                        if not object.hasProperty(property) :
                                                status = status + self.errormessage("Object %s has no property %s" % (self.ObjectPath(object), property))
                                        else :
                                                # in the following lines the absence of a _updateProperty
                                                # attribute indicates an object without properties (e.g. a method)
                                                # which indicates an object for which setting properties is a nonsense
                                                if hasattr(object, "_updateProperty") :
                                                        object._updateProperty(property, propvalue)
                                                        self.htmlmessage("Object %s property %s was modified to %s" % (self.ObjectPath(object), property, str(propvalue)))
                return status

        def run_addprop(self, expanded, unexpanded) :
                """Adds a property to objects

                   Accepts --name and --type mandatory options with arguments,
                   and an optional --value option with an argument.

                   Accepts multiple objects ids as its other arguments:

                   addprop --name description --type string .

                   This adds a property named 'description' and type 'string'
                   to the current Folder, leaving this property's value empty.

                   addprop --name author --type string --value Shakespeare /QuickStart/*_html

                   This adds a property named 'author' of type 'string'
                   and which value is 'Shakespeare' to all objects
                   which id match *_html in the /QuickStart Folder

                   Caveats: No check is done to verify that the property type is valid.
                """   
                options, args = self.getopt(["name=", "type=", "value="], expanded)
                if (options is None) and (args is None) :
                        return -1       # message was already displayed in self.getopt()
                if not args :
                        return self.errormessage("Needs at least one object to add a property to")
                if not (options.has_key("name") and options.has_key("type")) :
                        return self.errormessage("A property needs a name, a type and an optional value")
                property = options["name"]
                proptype = options["type"]
                if options.has_key("value") :
                        propvalue = options["value"]
                        try :
                                # maybe it's a list in a string, e.g. "['e', 'f']"
                                # or something like that
                                newvalue = eval(propvalue)
                                if (type(newvalue) != type(0)) and (type(newvalue) != type(0.0)) :
                                        # we mustn't convert numeric to string
                                        propvalue = newvalue
                        except NameError :
                                pass    # normal string
                        except SyntaxError :
                                pass    # string containing spaces
                else :
                        propvalue = ""
                status = 0
                for arg in args :
                        object = self.toObject(self.__context, arg)
                        if object is not None :
                                if not self.HasPerms(object, 'Manage properties') :
                                        status = status - 1
                                elif hasattr(object, 'hasProperty') :
                                        if object.hasProperty(property) :
                                                status = status + self.errormessage("Object %s already has property %s" % (self.ObjectPath(object), property))
                                        else :
                                                # in the following lines the absence of a _setProperty
                                                # attribute indicates an object without properties (e.g. a method)
                                                # which indicates an object for which setting properties is a nonsense
                                                if hasattr(object, "_setProperty") :
                                                        object._setProperty(property, propvalue, proptype)
                                                        self.htmlmessage("Property %s of type %s and value %s was added to object %s" % (property, proptype, str(propvalue), self.ObjectPath(object)))
                return status

        def run_delprop(self, expanded, unexpanded) :
                """Deletes objects properties

                   Accepts a property name as its first argument.
                   Accepts multiple objects ids as its other arguments:

                   delprop MyProperty MyFolder/* *html
                """   
                if len(expanded) < 2 :
                        return self.errormessage("Needs at least a property name and one object id")

                property = expanded[0]
                status = 0
                for arg in expanded[1:] :
                        object = self.toObject(self.__context, arg)
                        if object is not None :
                                if not self.HasPerms(object, 'Manage properties') :
                                        status = status - 1
                                elif hasattr(object, 'hasProperty') :
                                        if not object.hasProperty(property) :
                                                status = status + self.errormessage("Object %s has no property %s" % (self.ObjectPath(object), property))
                                        else :
                                                # in the following lines the absence of a _delProperty
                                                # attribute indicates an object without properties (e.g. a method)
                                                # which indicates an object for which deleting properties is a nonsense
                                                if hasattr(object, "_delProperty") :
                                                        object._delProperty(property)
                                                        self.htmlmessage("Property %s deleted from object %s" % (property, self.ObjectPath(object)))
                return status

        def run_lsprop(self, expanded, unexpanded) :
                """List objects properties

                   Accepts multiple arguments:

                   lsprop Contr?l_Pan*/Products/Sq* MyFolder/*

                   Caveats: Not sure that the result is correct wrt acquisition,
                            please report problems.
                """
                if not expanded :
                        return self.errormessage("Needs at least one object id")

                status = 0
                results = []
                for arg in expanded :
                        object = self.toObject(self.__context, arg)
                        if object is not None :
                                if not self.HasPerms(object, 'Access contents information') :
                                        status = status - 1
                                elif not hasattr(object, "propertyMap") :
                                        status = status + self.errormessage("Object %s doesn't have any property" % self.ObjectPath(object))
                                else :
                                        # in the following lines the absence of an _updateProperty
                                        # attribute indicates an object without properties (e.g. a method)
                                        # which indicates an object for which listing properties is a nonsense
                                        if hasattr(object, "_updateProperty") :
                                                for prop in object.propertyMap() :
                                                        propid = "%s.%s" % (object.getId(), prop["id"])
                                                        proptype = prop["type"]
                                                        propmode = prop.get("mode", "")
                                                        propvalue = repr(object.getProperty(prop["id"], 'Error'))

                                                        results.append({"Property": propid, "Type": proptype, "Value": propvalue, "Mode": propmode})
                                                        self.printf("%s\n" % propid)
                self.tableDisplay("lsprop", ["Property", "Type", "Value", "Mode"], results)
                return status


def zshell(self, zshellscript=None, xmlrpcstuff=None) :
    # so we can check for the type of data coming in to see
    # if this being called by xmlrpc
    from types import DictType

    # ok we need to be a little more helpful in what we are passing through
    # xmlrpc doesnot allow key = value args, so we will pass through a
    # hash for the moment and sort it all out.
    r = None
    if type(xmlrpcstuff)==DictType:
        self.REQUEST['zshellscript'] = xmlrpcstuff['command']
        self.REQUEST['zshellpwd'] = xmlrpcstuff['path']
        r = xmlrpcstuff['result_type']

    MyShell = ZopeShell(self)

    # well we dont want to return html to command line,
    # plus if you do a cd, we want to know the path.
    # so lets pass this back in a dict that xml rpc knows how to marshal
    if r == 'text':
        output = { 'data':MyShell.get_stdout(), 'path':MyShell.getcwd() }
        return output
    else :
        redirect = None
        if self.REQUEST.has_key("zshellredirect") and self.REQUEST["zshellredirect"] and string.strip(self.REQUEST["zshellredirect"]) :
            redirect = "%s/manage_main" % MyShell.getContext().absolute_url()
        return MyShell.get_HTML(redirect)
